# coding=utf-8
# Copyright 2023-present the International Business Machines.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Atomic fact decontextualization using LLMs

import json
import asyncio
import mellea.stdlib.functional as mfuncs

from typing import Any, Dict, List
from mellea.backends import Backend
from mellea.stdlib.context import SimpleContext
from mellea.stdlib.requirements import check, simple_validate
from mellea.stdlib.sampling import RejectionSamplingStrategy
from mellea.core import FancyLogger

# Local imports
from fact_reasoner.utils import validate_json_code_block, strip_code_fences, LOOP_BUDGET

INSTRUCTION_REVISER = """
Instructions:
You task is to decontextualize a UNIT to make it standalone. Each UNIT is an independent content unit or atomic unit extracted from the broader context of a RESPONSE.   

Vague References:
- Pronouns (e.g., "he", "she", "they", "it")
- Demonstrative pronouns (e.g., "this", "that", "these", "those")
- Unknown entities (e.g., "the event", "the research", "the invention")
- Incomplete names (e.g., "Jeff..." or "Bezos..." when referring to Jeff Bezos)

Follow the steps below for unit decontextualization:
1. If the UNIT contains vague references, minimally revise them with respect to the specific subjects they refer to in the RESPONSE.
2. The decontextualized UNIT should be minimally revised by ONLY resolving vague references. No additional information must be added.
3. UNIT extraction might decompose a conjunctive statement into multiple units (e.g. Democracy treats citizens as equals regardless of their race or religion -> (1) Democracy treats citizens as equals regardless of their race, (2) Democracy treats citizens as equals regardless of their religion). Avoid adding what is potentially part of another UNIT.
4. Provide a reasoning of the revisions you made to the UNIT, justifying each decision.
5. The output must be in the following JSON format with a markdown code block:

```json
{
  "revised_unit": "<REVISED_UNIT>",
  "rationale": "<YOUR_REASONING>"
}
```
Where <REVISED_UNIT> is the decontextualized UNIT after resolving vague references, and <YOUR_REASONING> is your reasoning for the revisions made.

Use the provided examples to learn your task.

Example 1: 
UNIT: 
Acorns is a financial technology company

RESPONSE:
Acorns is a financial technology company founded in 2012 by Walter Cruttenden, \
Jeff Cruttenden, and Mark Dru that provides micro-investing services. The \
company is headquartered in Irvine, California.

OUTPUT:
```json
{
    "revised_unit": "Acorns is a financial technology company.",
    "rationale": "This UNIT does not contain any vague references. Thus, the unit does not require any further decontextualization."
}
```

Example 2: 
UNIT:
The victim had previously suffered a broken wrist.

RESPONSE:
The clip shows the victim, with his arm in a cast, being dragged to the floor \
by his neck as his attacker says "I'll drown you" on a school playing field, while forcing water from a bottle into the victim's mouth, \
simulating waterboarding. The video was filmed in a lunch break. The clip shows the victim walking away, without reacting, as the attacker \
and others can be heard continuing to verbally abuse him. The victim, a Syrian refugee, had previously suffered a broken wrist; this had also been \
investigated by the police, who had interviewed three youths but took no further action.

OUTPUT:
```json
{
    "revised_unit": "The Syrian refugee victim had previously suffered a broken wrist.",
    "rationale": "The UNIT contains a vague reference, 'the victim.' This is a reference to an unknown entity, since it is unclear who the victim is. From the RESPONSE, we can see that the victim is a Syrian refugee. Thus, the vague reference 'the victim' should be replaced with 'the Syrian refugee victim.'"
}
```

Example 3:
UNIT:
The difference is relatively small.

RESPONSE:
Both the RTX 3060 Ti and RTX 3060 are powerful GPUs, and the difference between them lies in their performance. \
The RTX 3060 Ti has more CUDA cores (4864 vs 3584) but a lower boost clock speed (1665 MHz vs 1777 MHz) compared to the RTX 3060. \
In terms of memory bandwidth, the RTX 3060 Ti has a slight edge over the RTX 3060 with a bandwidth of 448 GB/s compared to 360 GB/s. \
However, the difference is relatively small and may not be noticeable in real-world applications.

OUTPUT:
```json
{
    "revised_unit": "The difference in memory bandwidth between the RTX 3060 Ti and RTX 3060 is relatively small.",
    "rationale": "The UNIT contains a vague reference, 'The difference.' From the RESPONSE, we can see that the difference is in memory bandwidth between the RTX 3060 Ti and RTX 3060. Thus, the vague reference 'The difference' should be replaced with 'The difference in memory bandwidth between the RTX 3060 Ti and RTX 3060'. The sentence from which the UNIT is extracted includes coordinating conjunctions that potentially decompose the statement into multiple units. Thus, adding more context to the UNIT is not necessary."
}
```

Your task:
UNIT:
{{atomic_unit}}

RESPONSE:
{{response}}

OUTPUT:
"""

class Reviser:
    """
    Atomic unit decontextualization using LLMs.
    
    """

    def __init__(
            self,
            backend: Backend,
    ):
        """
        Initialize the Reviser.

        Args:
            backend: Backend
                The Mellea backend to use for LLM interactions.
        """ 
        
        # Safety checks        
        if backend is None:
            raise ValueError("Mellea session is None. Please provide a valid Mellea session.")

        # Initialize the reviser
        self.backend = backend
        
        # Print backend info
        print(f"[Reviser] Using Mellea backend: {self.backend.model_id}")

        # Disable Mellea logging
        FancyLogger.get_logger().setLevel(FancyLogger.ERROR)


    def run(self, units: List[str], response: str) -> List[Dict[str, Any]]:
        """
        Decontextualize the input atomic units using the response as context.
        
        Args:
            units: List[str]
                The atomic units to be decontextualized.
            response: str
                The response from which the atomic unit is decontextualized.
        Returns:
            List[str]: A dictionary containing the revised atomic unit.
        """

        # Perform the instruction with validation
        results = []
        for atom_text in units:
            output = mfuncs.instruct(
                INSTRUCTION_REVISER,
                context=SimpleContext(),
                backend=self.backend,
                requirements=[
                    check(
                        "The output must be a valid JSON code block.",
                        validation_fn=simple_validate(
                            lambda s: validate_json_code_block(s, required_keys=["revised_unit", "rationale"])
                        )
                    )
                ],
                user_variables={"atomic_unit": atom_text, "response": response},
                strategy=RejectionSamplingStrategy(loop_budget=LOOP_BUDGET),
                return_sampling_results=True
            )

            if output.success:
                cleaned = strip_code_fences(str(output))
                revised_unit = json.loads(cleaned)
                revised_unit.update({"text": atom_text})
                results.append(revised_unit)
        
        return results

    async def run_batch(self, units: List[str], response: str) -> List[Dict[str, Any]]:
        """
        Decontextualize the input atomic units using the response as context.
        
        Args:
            units: List[str]
                The atomic units to be decontextualized.
            response: str
                The response from which the atomic unit is decontextualized.
        Returns:
            List[str]: A dictionary containing the revised atomic unit.
        """

        # Perform the instruction with validation
        
        corutines = []
        for atom_text in units:
            corutine = mfuncs.ainstruct(
                INSTRUCTION_REVISER,
                context=SimpleContext(),
                backend=self.backend,
                requirements=[
                    check(
                        "The output must be a valid JSON code block.",
                        validation_fn=simple_validate(
                            lambda s: validate_json_code_block(s, required_keys=["revised_unit", "rationale"])
                        )
                    )
                ],
                user_variables={"atomic_unit": atom_text, "response": response},
                strategy=RejectionSamplingStrategy(loop_budget=LOOP_BUDGET),
                return_sampling_results=True
            )
            corutines.append(corutine)

        results = []
        print(f"[Reviser] Awaiting for async execution ...")
        outputs = await asyncio.gather(*(corutines[i] for i in range(len(corutines))))
        for output in outputs:
            if output.success:
                cleaned = strip_code_fences(str(output))
                revised_unit = json.loads(cleaned)
                revised_unit.update({"text": atom_text})
                results.append(revised_unit)
        
        return results
    



# coding=utf-8
# Copyright 2023-present the International Business Machines.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Query builder for atoms to retrieve results from Google and/or Wikipedia

import mellea.stdlib.functional as mfuncs

from mellea.backends import Backend
from mellea.stdlib.context import SimpleContext
from mellea.stdlib.requirements import check, simple_validate
from mellea.stdlib.sampling import RejectionSamplingStrategy
from mellea.core import FancyLogger

# Local imports
from fact_reasoner.utils import validate_markdown_code_block, strip_code_fences

INSTRUCTION_QUERY_BUILDER = """
Instructions:
Your task is to generate a Google Search query about a given STATEMENT. Your goal is to create a high-quality query that is most likely to retrieve relevant information about the STATEMENT.

QUERY CONSTRUCTION CRITERIA:
A well-crafted query should:
  - Retrieve information to verify the STATEMENT's factual accuracy.
  - Balance specificity for targeted results with breadth to avoid missing critical information.

Process:
1. Construct a Useful Google Search Query:
   - Craft a query based on the QUERY CONSTRUCTION CRITERIA.
   - Prioritize natural language queries that a typical user might enter.
   - Use special operators (quotation marks, "site:", Boolean operators, intitle:, etc.) selectively and only when they significantly enhance the query's effectiveness.
   
2. Format Final Query:
   Present your query wrapped between Markdown code fences:
   ```
    <your generated query here>
   ```

Use the following examples to learn the task better.

Example 1:
STATEMENT: The Great Wall of China is visible from space
OUTPUT:
```
"The Great Wall of China is visible from space" fact check myth
```

Example 2:
STATEMENT: Apple will release a foldable iPhone in 2026
OUTPUT:
```
"Apple will release a foldable iPhone in 2026" rumor OR announcement
```

Example 3:
STATEMENT: Quantum computers can break RSA encryption easily
OUTPUT:
```
"Quantum computers can break RSA encryption easily" fact check cryptography experts
```

Your task:

STATEMENT: {{statement_text}}
OUTPUT:
"""

class QueryBuilder:
    """
    The QueryBuilder uses an LLM to generate a query string. The query is then
    used to retrieve results from Google Search, Wikipedia, ChromaDB.
    """

    def __init__(
        self, 
        backend: Backend
    ):
        """
        Initialize the QueryBuilder.

        Args:
            backend: Backend
                The Mellea backend to use for LLM interaction.
        """

        # Safety checks        
        if backend is None:
            raise ValueError("Mellea backend is None. Please provide a valid Mellea backend.")

        # Initialize the extractor
        self.backend = backend
        
        # Print info
        print(f"[QueryBuilder] Using Mellea backend: {self.backend.model_id}")

        # Disable Mellea logging
        FancyLogger.get_logger().setLevel(FancyLogger.ERROR)


    def run(self, text: str) -> str:
        """
        Build a Google search query for the given text.
        
        Args:
            text: str
                The text for which to build the Google search query.
        Returns:
            dict: A dictionary containing the query string.
        """
        
        # Perform the instruction with validation
        output = mfuncs.instruct(
            INSTRUCTION_QUERY_BUILDER,
            context=SimpleContext(),
            backend=self.backend,
            requirements=[
                check(
                    "The output must be wrapped with markdown code fences.",
                    validation_fn=simple_validate(
                        lambda s: validate_markdown_code_block(s)
                    ),
                )
            ],
            user_variables={"statement_text": text},
            strategy=RejectionSamplingStrategy(loop_budget=3),
            return_sampling_results=True,
        )

        # The output is a validated JSON string; parse it
        if output.success:
            cleaned = strip_code_fences(str(output))
            return cleaned
        else:
            return text # the original text
                        



# coding=utf-8
# Copyright 2023-present the International Business Machines.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Decompose the input string into atomic units. Use the same Mellea session 
# (context) to revise or decontextualize the atomc units if needed.

import json
import asyncio
import mellea.stdlib.functional as mfuncs

from typing import Dict, List
from mellea.backends import Backend
from mellea.stdlib.context import SimpleContext
from mellea.stdlib.requirements import check, simple_validate
from mellea.stdlib.sampling import RejectionSamplingStrategy
from mellea.core import FancyLogger

# Local imports
from fact_reasoner.utils import validate_json_code_block, strip_code_fences, LOOP_BUDGET

INSTRUCTION_ATOMIZER = """
Instructions:
Your task is to break down a given paragraph into a set of atomic units without adding any new information.

Rules:
- An atomic unit is the smallest sentence containing a singular piece of information directly extracted from the provided paragraph.
- Atomic units may contradict one another.
- The paragraph may contain information that is factually incorrect. Even in such cases, you are not to alter any information contained in the paragraph and must produce atomic units that are completely faithful to the information in the paragraph.
- Each atomic unit in the output must check a different piece of information found explicitly in the paragraph.
- Each atomic unit is standalone in that any actual nouns or proper nouns should be used in place of pronouns or anaphors.
- Each atomic unit must not include any information beyond what is explicitly stated in the provided paragraph.
- Where possible, avoid paraphrasing and instead try to only use language used in the paragraph without introducing new words. 
- The output must be a JSON dictionary with the following format and markdown code fences such that each atomic unit has a unique ID:

```json
{
    "id1": "<first atomic unit>",
    "id2": "<second atomic unit>",
    ...
}
```

Use the provided examples to learn your task.

Example 1:
INPUT: Glenn Allen Anzalone (born June 23, 1955), better known by his stage name Glenn Danzig, is an American singer, songwriter, musician, and record producer. He is the founder of the rock bands Misfits, Samhain, and Danzig. He owns the Evilive record label as well as Verotik, an adult-oriented comic book publishing company.
OUTPUT:
```json
{
    "id1": "Glenn Allen Anzalone was born on June 23, 1955.",
    "id2": "Glenn Allen Anzalone is better known by his stage name Glenn Danzig.",
    "id3": "Glenn Danzig is an American singer, songwriter, musician, and record producer.",
    "id4": "Glenn Danzig is the founder of several rock bands, including Misfits, Samhain, and Danzig.",
    "id5": "Glenn Danzig owns the Evilive record label.",
    "id6": "Glenn Danzig owns Verotik, which is an adult-oriented comic book publishing company."
}
```

Example 2:
INPUT: Luiz Inácio Lula da Silva (born 27 October 1945), also known as Lula da Silva or simply Lula, is a Brazilian politician who is the 39th and current president of Brazil since 2023. A member of the Workers' Party, Lula was also the 35th president from 2003 to 2010. He also holds the presidency of the G20 since 2023. Lula quit school after second grade to work, and did not learn to read until he was ten years old. As a teenager, he worked as a metalworker and became a trade unionist.
OUTPUT:
```json
{
    "id1": "Luiz Inácio Lula da Silva was born on October 27, 1945.",
    "id2": "Luiz Inácio Lula da Silva is also known as Lula da Silva or simply Lula.",
    "id3": "Lula is a Brazilian politician.",
    "id4": "Lula is the 39th and current president of Brazil since 2023.",
    "id5": "Lula is a member of the Workers' Party.",
    "id6": "Lula served as the 35th president of Brazil from 2003 to 2010.",
    "id7": "Lula holds the presidency of the G20 since 2023.",
    "id8": "Lula quit school after the second grade to work.",
    "id9": "Lula did not learn to read until he was ten years old.",
    "id10": "As a teenager, Lula worked as a metalworker.",
    "id11": "Lula became a trade unionist."
}
```

Example 3:
INPUT: Zhejiang Huafang Pharmaceutical Co., Ltd. is a leading chemical company based in China that specializes in the research, manufacturing, and sales of various pharmaceutical products, including excipients and intermediates. The company was founded in 2018 and is located in Hangzhou, a city with a rich history in eastern China. Zhejiang Huafang Pharmaceutical Co., Ltd. is committed to providing high-quality products to its customers in the healthcare industry. The company's manufacturing facilities are equipped with state-of-the-art technology and infrastructure that ensure the production of high-quality products. Overall, Zhejiang Huafang Pharmaceutical Co., Ltd. is a reputable pharmaceutical company with a long history of success in the healthcare industry. The company's commitment to quality, innovation, and customer service has made it a leader in the field of pharmaceutical research and development.
OUTPUT:
```json
{
    "id1": "Zhejiang Huafang Pharmaceutical Co., Ltd. is a leading chemical company.",
    "id2": "Zhejiang Huafang Pharmaceutical Co., Ltd. is based in China.",
    "id3": "Zhejiang Huafang Pharmaceutical Co., Ltd. specializes in the research of various pharmaceutical products",
    "id4": "Zhejiang Huafang Pharmaceutical Co., Ltd. specializes in the manufacturing of various pharmaceutical products.",
    "id5": "Zhejiang Huafang Pharmaceutical Co., Ltd. specializes in the sales of various pharmaceutical products.",
    "id6": "Excipients are the pharmaceutical products of the Zhejiang Huafang Pharmaceutical Co., Ltd.",
    "id7": "Intermediates are the pharmaceutical products of the Zhejiang Huafang Pharmaceutical Co., Ltd.",
    "id8": "The company was founded in 2018.",
    "id9": "The company is located in Hangzhou.",
    "id10": "Hangzhou is a city.",
    "id11": "Hangzhou has a rich history in eastern China.",
    "id12": "Zhejiang Huafang Pharmaceutical Co., Ltd. is committed to providing high-quality products to its customers in the healthcare industry.",
    "id13": "The company's manufacturing facilities are equipped with state-of-the-art technology.",
    "id14": "The company's manufacturing facilities are equipped with state-of-the-art infrastructure.",
    "id15": "The company's manufacturing facilities are equipped with state-of-the-art technology and infrastructure that ensure the production of high-quality products.",
    "id16": "Zhejiang Huafang Pharmaceutical Co., Ltd. is a reputable pharmaceutical company.",
    "id17": "Zhejiang Huafang Pharmaceutical Co., Ltd. has a long history of success in the healthcare industry.",
    "id18": "The company is committed to quality.",
    "id19": "The company is committed to innovation.",
    "id20": "The company is committed to customer service.",
    "id21": "The company's commitment to quality, innovation, and customer service has made it a leader in the field of pharmaceutical research.",
    "id22": "The company's commitment to quality, innovation, and customer service has made it a leader in the field of pharmaceutical development."
}
```

Example 4:
INPUT: I'm here to help you make an informed decision. Both the RTX 3060 Ti and RTX 3060 are powerful GPUs, and the difference between them lies in their performance. The RTX 3060 Ti has more CUDA cores (4864 vs 3584) but a lower boost clock speed (1665 MHz vs 1777 MHz) compared to the RTX 3060. In terms of memory bandwidth, the RTX 3060 Ti has a slight edge over the RTX 3060 with a bandwidth of 448 GB/s compared to 360 GB/s. However, the difference is relatively small. It's important to consider other factors such as the power consumption, cooling system, and compatibility with your system when making a decision."
OUTPUT:
```json
{
    "id1": "The RTX 3060 Ti is a powerful GPU.",
    "id2": "The RTX 3060 is a powerful GPU.",
    "id3": "The difference between them lies in their performance.",
    "id4": "The RTX 3060 Ti has more CUDA cores compared to the RTX 3060.",
    "id5": "The RTX 3060 Ti has 4864 CUDA cores.",
    "id6": "The RTX 3060 has 3584 CUDA cores.",
    "id7": "The RTX 3060 Ti has a lower boost clock speed compared to the RTX 3060.",
    "id8": "The RTX 3060 Ti has a boost clock speed of 1665 MHz.",
    "id9": "The RTX 3060 has a boost clock speed of 1777 MHz.",
    "id10": "The RTX 3060 Ti has a slight edge over the RTX 3060 in terms of memory bandwidth.",
    "id11": "The RTX 3060 Ti has a memory bandwidth of 448 GB/s.",
    "id12": "The RTX 3060 has a memory bandwidth of 360 GB/s.",
    "id13": "The difference is relatively small.",
}
```

Your task:
INPUT: {{response}}
OUTPUT:
"""

class Atomizer(object):
    """
    The Atomizer class implements the atomic decomposition of the response.
    For our purpose, an atomic unit or atom is either a fact or a claim.
    """

    def __init__(
        self,
        backend: Backend,
    ):
        """
        Initialize the Atomizer.

        Args:
            backend: Backend
                The Mellea backend to use for LLM interactions.
        """ 

        # Safety checks        
        if backend is None:
            raise ValueError("Mellea backend is None. Please provide a valid Mellea backend.")

        # Initialize the extractor
        self.backend = backend
        
        # Print info
        print(f"[Atomizer] Using Mellea backend: {self.backend.model_id}")

        # Disable Mellea logging
        FancyLogger.get_logger().setLevel(FancyLogger.ERROR)

    def run(self, response: str) -> Dict[str, str]:
        """
        Extract atomic units from a single response.
        
        Args:
            response: str
                The response from which to extract atomic units.
        Returns:
            Dict[str, str]: A dictionary containing the atomic units, each with
            a unique identifier.
        """
        
        # Perform the instruction with validation
        output = mfuncs.instruct(
            INSTRUCTION_ATOMIZER,
            context=SimpleContext(),
            backend=self.backend,
            requirements=[
                check(
                    "The output must be a valid JSON dictionary with markdown code fences",
                    validation_fn=simple_validate(
                        lambda s: validate_json_code_block(s)
                    ),
                )
            ],
            user_variables={"response": response},
            strategy=RejectionSamplingStrategy(loop_budget=LOOP_BUDGET),
            return_sampling_results=True,
        )

        # The output is a validated JSON string; parse it
        if output.success:
            cleaned = strip_code_fences(str(output))
            return json.loads(cleaned)
        else:
            return {} # empty dict on failure
        
    async def run_batch(self, responses: List[str]) -> List[Dict[str, str]]:
        """
        Extract atomic units from a list of responses.
        
        Args:
            responses: List[str]
                The list of response from which to extract atomic units.
        Returns:
            dict: A dictionary containing the number of atomic units, the units themselves,
            all atomic units as dictionaries, and all facts as dictionaries.
        """
        
        # Perform the instruction with validation
        corutines = []
        for response in responses:
            corutine = mfuncs.ainstruct(
                INSTRUCTION_ATOMIZER,
                context=SimpleContext(),
                backend=self.backend,
                requirements=[
                    check(
                        "The output must be a valid JSON dictionary with markdown code fences",
                        validation_fn=simple_validate(
                            lambda s: validate_json_code_block(s)
                        ),
                    )
                ],
                user_variables={"response": response},
                strategy=RejectionSamplingStrategy(loop_budget=LOOP_BUDGET),
                return_sampling_results=True,
            )
            corutines.append(corutine)

        results = []
        print(f"[Atomizer] Awaiting for the async execution ...")
        outputs = await asyncio.gather(*(corutines[i] for i in range(len(corutines))))
        for output in outputs:
            # The output is a validated JSON string; parse it
            if output.success:
                cleaned = strip_code_fences(str(output))
                results.append(json.loads(cleaned))
            else:
                results.append({}) # empty dict on failure

        return results

    def __str__(self) -> str:
        return "This is the atomizer"    



# coding=utf-8
# Copyright 2023-present the International Business Machines.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import numpy as np
import requests
import tqdm
import os
import re
import random
import torch
import transformers

from typing import List, Union

LOOP_BUDGET = 5
class dotdict(dict):
    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

# GPU related utils    
def get_freer_gpu():
    os.system('nvidia-smi -q -d Memory |grep -A6 GPU|grep Free >tmp_smi')
    memory_available = [int(x.split()[2])+5*i for i, x in enumerate(open('tmp_smi', 'r').readlines())]
    os.remove("tmp_smi")
    return np.argmax(memory_available)

def select_freer_gpu():
    freer_gpu = str(get_freer_gpu())
    print("Will use GPU: %s" % (freer_gpu))
    os.environ['CUDA_LAUNCH_BLOCKING'] = "1"
    os.environ["CUDA_VISIBLE_DEVICES"] = ""+freer_gpu
    return freer_gpu

# Set the random seed globally
def set_seed(seed: int):
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    transformers.set_seed(seed)

# String manipulation utils
def join_segments(*args: Union[str, List[str]], separator: str = '\n\n\n') -> str:
    """Joins an unspecified number of strings using the separator."""
    all_segments = []

    for arg in args:
        if isinstance(arg, list):
            all_segments.extend(arg)
        else:
            all_segments.append(strip_string(str(arg)))

    return strip_string(separator.join(all_segments))


def strip_string(s: str) -> str:
    """Strips a string of newlines and spaces."""
    return s.strip(' \n')

def punctuation_only_inside_quotes(text):
    # find all quoted sections (single or double quotes)
    quoted_spans = [match.span() for match in re.finditer(r'"[^"]*"|\'[^\']*\'', text)]

    def is_inside_quotes(index):
        return any(start < index < end for start, end in quoted_spans)

    # check each comma and semicolon
    for i, char in enumerate(text):
        if char in [",", ";"]:
            if not is_inside_quotes(i):
                return False  # found punctuation outside quotes
    return True


def extract_first_square_brackets(input_string: str) -> str:
    """Extracts the contents of the FIRST string between square brackets."""
    raw_result = re.findall(r'\[.*?\]', input_string, flags=re.DOTALL)

    if raw_result:
        return raw_result[0][1:-1]
    else:
        return ''

def extract_last_square_brackets(input_string: str) -> str:
    """Extracts the contents of the FIRST string between square brackets."""
    raw_result = re.findall(r'\[.*?\]', input_string, flags=re.DOTALL)

    if raw_result:
        return raw_result[-1][1:-1]
    else:
        return ''

def extract_last_wrapped_response(input_string: str) -> str:
    """Extracts the contents of the LAST string between pairs of ###."""
    raw_result = re.findall(r'###.*?###', input_string, flags=re.DOTALL)

    if raw_result:
        return raw_result[-1][3:-3]
    else:
        return ''


def extract_first_code_block(
    input_string: str, ignore_language: bool = False
) -> str:
    """Extracts the contents of a string between the first code block (```)."""
    if ignore_language:
        pattern = re.compile(r'```(?:\w+\n)?(.*?)```', re.DOTALL)
    else:
        pattern = re.compile(r'```(.*?)```', re.DOTALL)

    match = pattern.search(input_string)
    return strip_string(match.group(1)) if match else ''


def batcher(iterator, batch_size=4, progress=False):
    if progress:
        iterator = tqdm.tqdm(iterator)

    batch = []
    for elem in iterator:
        batch.append(elem)
        if len(batch) == batch_size:
            final_batch = batch
            batch = []
            yield final_batch
    if len(batch) > 0: # Leftovers
        yield batch

# Google Drive related

def download_file_from_google_drive(id, destination):
    URL = "https://docs.google.com/uc?export=download"

    session = requests.Session()

    response = session.get(URL, params = { 'id' : id }, stream = True)
    token = get_confirm_token(response)

    if token:
        params = { 'id' : id, 'confirm' : token }
        response = session.get(URL, params = params, stream = True)

    save_response_content(response, destination)    

def get_confirm_token(response):
    for key, value in response.cookies.items():
        if key.startswith('download_warning'):
            return value

    return None

def save_response_content(response, destination):
    CHUNK_SIZE = 32768

    with open(destination, "wb") as f:
        for chunk in response.iter_content(CHUNK_SIZE):
            if chunk: # filter out keep-alive new chunks
                f.write(chunk)

def strip_code_fences(s: str) -> str:
    """
    Strip markdown code fences from a string if present.

    Args:
        s: str 
            The input string.
    Returns:
        str: The string without code fences.
    """

    s = s.strip()

    # Try a strict fenced block: ```json\n ... \n```
    m = re.match(r"^```(?:json|JSON)?\s*\n(.*?)\n```$", s, re.DOTALL)
    if m:
        return m.group(1).strip()

    # Fallback: starts with ``` but may have irregular spacing/line breaks
    if s.startswith("```"):
        lines = s.splitlines()
        # Remove the opening fence line
        content_lines = lines[1:]
        # If the last line is a closing fence, drop it
        if content_lines and content_lines[-1].strip().startswith("```"):
            content_lines = content_lines[:-1]
        return "\n".join(content_lines).strip()

    # No fences detected; return as-is
    return s

def normalize_ws(text: str) -> str:
    """
    - Collapse all runs of whitespace (spaces, tabs, newlines) into a single space.
    - Escape inner pairs of double quotes:  ""phrase""  ->  \"phrase\"
      (Only when the pair of quotes is not already escaped.)

    Examples:
        Input:  'He said  ""Hello  world"" \n  today.'
        Output: 'He said \\"Hello world\\" today.'
    """
    if text is None:
        return text  # or raise ValueError("text must not be None")

    # Collapse all whitespace chunks to a single space
    # This turns tabs/newlines into spaces as well.
    collapsed = re.sub(r'\s+', ' ', text).strip()
    collapsed = collapsed.replace('\n', '')

    return collapsed


def validate_json_code_block(input_string: str, required_keys: List[str] = None) -> bool:
    """
    Checks if the input string is a valid JSON dictionary.

    Args:
        input_string: str
            The string to check.
        required_keys: List[str]
            List of keys that must be present in the JSON dictionary.

    Returns:
        bool: True if valid JSON, False otherwise. If required_keys is provided,
        then also checks if it is a dictionary and contains the required keys.
    """
    try:

        # Remove markdown fences if present
        cleaned = strip_code_fences(input_string)
        cleaned = normalize_ws(cleaned)
        
        # Attempt to parse the string as JSON
        data = json.loads(cleaned)

        # Check if it's a dictionary and has required keys
        if isinstance(data, dict) and required_keys:
            for key in required_keys:
                if key not in data:
                    return False
        return True
    except json.JSONDecodeError as e:
        # If parsing fails, it's not valid JSON
        print(f"Malformed JSON string: {e}")
        return False

def validate_markdown_code_block(input_string: str) -> bool:
    """
    Checks if the input string is a valid markdown code block.

    Args:
        input_string: str
            The string to check.

    Returns:
        bool: True if valid markdown code block, False otherwise.
    """

    if input_string.strip().startswith("```") and input_string.strip().endswith("```"):
        return True
    else:
        return False
    


# coding=utf-8
# Copyright 2023-present the International Business Machines.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Factuality graph

import json
from pathlib import Path
from typing import List
from tqdm import tqdm
import networkx as nx

class Node:
    def __init__(
            self, 
            id: str, 
            type: str,
            probability: float = 1.0
    ):
        """
        Create a node in the graph.

        Args:
            id: str
                Unique ID of the node in the graph.
            type: str
                The node type: ["atom", "context"].
            probability: float
                The prior probability associated with the "atom" or "context".
        """

        assert (type in ["atom", "context"]), \
            f"Uknown node type: {type}."
        self.id = id
        self.type = type
        self.probability = probability

    def __str__(self):
        return f"Node {self.id} ({self.type}): {self.probability}"
    
class Edge:
    def __init__(
            self, 
            source: str, 
            target: str, 
            type: str, 
            probability: float,
            link: str
    ):
        """
        Create an edge in the graph.

        Args:
            source: str
                The `from` node ID in the graph.
            target: str
                The `to` node ID in the graph.
            type: str
                The NLI relation type represented by the edge. Allowed values are:
                ["entailment", "contradiction", "equivalence"].
            probability: float
                The probability value associated with the NLI relation type.
            link: str
                The type of link: context_atom, context_context, atom_atom
        """

        assert (type in ["entailment", "contradiction", "equivalence"]), \
            f"Unknown relation type: {type}."
        assert (link in ["context_atom", "context_context", "atom_atom"]), \
            f"Unknown link type: {link}"
        self.source = source
        self.target = target
        self.type = type
        self.probability = probability
        self.link = link

    def __str__(self):
        return f"[{self.source} -> {self.target} ({self.type}): {self.probability}]"
    
class FactGraph:
    """
    A graph representation of the atom-context relations.

    """

    def __init__(
            self,
            atoms: List = None,
            contexts: List = None,
            relations: List = None,
    ):
        """
        FactGraph constructor.

        Args:
            atoms: List
                The list of atoms in the answer.
            contexts: List
                The list of contexts for each of the atoms. Each context contains
                a reference to its corresponding atom (by construction).
            relations: List
                The list of relations between atoms and contexts.
        """

        # Initialize an empty graph
        self.nodes = {}
        self.edges = []

        if atoms is not None:
            for atom in tqdm(atoms, desc="Atoms"):
                node = Node(
                    id=atom.id, 
                    type="atom", 
                    probability=atom.probability
                )
                self.add_node(node)
        if contexts is not None:
            for context in tqdm(contexts, desc="Contexts"):
                node = Node(
                    id=context.id, 
                    type="context", 
                    probability=context.probability
                )
                self.add_node(node)

        if relations is not None:
            for rel in tqdm(relations, desc="Relations"):
                self.edges.append(
                    Edge(
                        source=rel.source.id,
                        target=rel.target.id,
                        type=rel.type,
                        probability=rel.probability,
                        link=rel.link
                    )
                )
        
    def get_nodes(self) -> list:
        return list(self.nodes.values())
    
    def get_edges(self) -> list:
        return self.edges
    
    def add_node(
            self, 
            node: Node
    ):
        """
        Add a new node to the graph.

        Args:
            node: Node
                A new node to be added to the graph
        """

        self.nodes[node.id] = node

    def add_edge(
            self,
            edge: Edge
    ):
        """
        Add a new edge to the graph.

        Args:
            edge: Edge
                A new edge to be added to the graph
        """

        self.edges.append(edge)

    def from_json(
            self,
            json_file: str
    ):
        """
        Create the FactGraph from a json file.

        Args:
            json_file: str
                Path to the json file containing the graph.
        """

        with open(json_file, "r") as f:
            data = json.load(f)
            assert ("nodes" in data and "edges" in data), f"Uknown graph format"

            self.nodes = {}
            for node in tqdm(data["nodes"], desc="Nodes"):
                self.add_node(
                    Node(
                        id=node["id"], 
                        type=node["type"],
                        probability=node["probability"]
                    )
                )
            
            self.edges = []
            for edge in tqdm(data["edges"], desc="Edges"):
                self.edges.append(
                    Edge(
                        source=edge["from"],
                        target=edge["to"],
                        type=edge["relation"],
                        probability=edge["probability"],
                        link=edge["link"]
                    )
                )

    def as_digraph(self):
        """
        Generate a networkx.DiGraph representation of the fact graph.
        """
        G = nx.DiGraph()
        for _, node in self.nodes.items():
            if node.type == "atom":
                G.add_node(node.id, color="green")
            else:
                G.add_node(node.id, color="orange")

        for edge in self.edges:
            if edge.type == "entailment":
                G.add_edge(edge.source, edge.target, color="green", label="{:.4g}".format(edge.probability))
            elif edge.type == "contradiction":
                G.add_edge(edge.source, edge.target, color="red", label="{:.4g}".format(edge.probability))
            elif edge.type == "equivalence":
                G.add_edge(edge.source, edge.target, color="blue", label="{:.4g}".format(edge.probability))

        return G


    def dump(self):
        print("Nodes:")
        for i, n in self.nodes.items():
            print(n)
        print("Edges:")
        for e in self.edges:
            print(e)
        print(f"Number of nodes: {len(self.nodes)}")
        print(f"Number of edges: {len(self.edges)}")

if __name__ == "__main__":

    d = Path(__file__).resolve().parent.parent.parent
    filename = Path.joinpath(d, "examples", "simple.json")
    fg = FactGraph()
    fg.from_json(json_file=filename)
    fg.dump()
    print(f"Done.")



# coding=utf-8
# Copyright 2023-present the International Business Machines.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Our implementation of the FactScore paper using LLAMA3 models

import json
import asyncio
import string
import time
import mellea.stdlib.functional as mfuncs

from typing import List, Dict, Any, Tuple
from mellea.backends import Backend
from mellea.stdlib.context import SimpleContext
from mellea.core import ModelOutputThunk
from mellea.stdlib.requirements import check
from mellea.stdlib.sampling import RejectionSamplingStrategy
from mellea.core import FancyLogger

# Local imports
from fact_reasoner.core.atomizer import Atomizer
from fact_reasoner.core.reviser import Reviser
from fact_reasoner.core.retriever import ContextRetriever
from fact_reasoner.core.base import Atom, Context
from fact_reasoner.core.utils import build_atoms, build_contexts, remove_duplicated_atoms
from fact_reasoner.utils import LOOP_BUDGET

# Version 1 of the prompt (from the original FactScore paper)
INSTRUCTION_FACTSCORE = """
Answer the question about {{topic_text}} based on the given context.
Your answer must be either True or False. Do not add any other information.

{{knowledge_text}}

Input: {{atom_text}} True or False?
Output:
"""

INSTRUCTION_FACTSCORE_NOTOPIC = """
Answer the input question based on the given context.
Your answer must be either True or False. Do not add any other information.

{{knowledge_text}}

Input: {{atom_text}} True or False?
Output:
"""

class FactScore:
    """
    Implementation of the FactScore factuality assessor.

    Source:
        @misc{min2023factscore,
        title={FActScore: Fine-grained Atomic Evaluation of Factual Precision in Long Form Text Generation}, 
        author={Sewon Min and Kalpesh Krishna and Xinxi Lyu and Mike Lewis and Wen-tau Yih and Pang Wei Koh and Mohit Iyyer and Luke Zettlemoyer and Hannaneh Hajishirzi},
        year={2023},
        eprint={2305.14251},
        archivePrefix={arXiv},
        primaryClass={cs.CL},
        url={https://arxiv.org/abs/2305.14251}, 
        }
    """

    def __init__(
            self,
            backend: Backend,
            atom_extractor: Atomizer = None,
            atom_reviser: Reviser = None,
            context_retriever: ContextRetriever = None,
    ):
        """
        Initialize the FactScore pipeline.

        Args:
            backend: Backend
                The Mellea backend to use for LLM interactions.
            atom_extractor: Atomizer
                The atom decomposition component.
            atom_reviser: Reviser
                The atom reviser component.
            context_retriever: ContextRetriever
                The service used for retrieving external contexts.
        """

        self.backend = backend
        self.query = None
        self.response = None
        self.topic = None
        self.start_time = time.perf_counter() # get the start time

        self.context_retriever = context_retriever
        self.atom_extractor = atom_extractor
        self.atom_reviser = atom_reviser
        self.binary_output = True # default is True
    
        print(f"[FactScore] Using Mellea backend: {self.backend.model_id}")
        print(f"[FactScore] Binary output: {self.binary_output}")

        self.atoms = {} # indexed by atom id
        self.contexts = {} # indexed by context id

        # Ground truth labels (if any)
        self.labels_human = None

        # Disable Mellea logging
        FancyLogger.get_logger().setLevel(FancyLogger.ERROR)

    def from_dict_with_contexts(
            self,
            data: Dict[str, Any],
    ):
        """
        Initialize FactScore from a dict containing both atoms and contexts.

        Args:
            data: Dict[str, Any]
                The dict containing the problem instance.
        """

        self.query = data["input"]
        self.response = data["output"]
        self.topic = data.get("topic", None)
        
        print(f"[FactScore] Reading the atoms ...")                
        gold_labels = []
        atom_ids = []
        self.atoms = {}
        self.contexts = {}
        atom2contexts = {}
        for atom_dict in data["atoms"]:
            aid = atom_dict["id"]
            text = atom_dict["text"]
            original = atom_dict["original"]
            label = atom_dict.get("label", None)
            contexts = atom_dict["contexts"]
            a = Atom(id=aid, text=text, label=label)
            a.set_original(original)
            atom_ids.append(aid)
            gold_labels.append(label)
            self.atoms[aid] = a
            atom2contexts[aid] = contexts

        print(f"[FactScore] Atoms found: {len(self.atoms)}")
        for _, atom in self.atoms.items():
            print(f"[FactScore] {atom}")
        
        self.labels_human = dict(zip(atom_ids, gold_labels))
        print(f"[FactScore] Lables found: {self.labels_human}")

        print(f"[FactScore] Reading the contexts ...")
        for context_dict in data["contexts"]:
            cid = context_dict["id"]
            title = context_dict["title"]
            text = context_dict["text"]
            snippet = context_dict.get("snippet", "")
            link = context_dict.get("link", "")
            ctxt = Context(
                id=cid, 
                atom=None, 
                text=text, 
                title=title, 
                snippet=snippet, 
                link=link
            )
            self.contexts[cid] = ctxt

        print(f"[FactScore] Contexts found: {len(self.contexts)}")
        for aid, atom in self.atoms.items():
            ctxts = []
            for c in atom2contexts[aid]:
                ctxts.append(self.contexts[c])
                self.contexts[c].set_atom(atom)
            atom.add_contexts(ctxts)
        print(f"[FactScore] Pipeline initialized with {len(self.atoms)} atoms and {len(self.contexts)} contexts.")

    def to_json(self, json_file_path: str = None) -> Dict[str, Any]:
        """
        Save the FactScore instance to a JSON file.

        Args:
            json_file: str
                The path to the output JSON file.
        """

        data = {}
        data["input"] = self.query
        data["output"] = self.response.strip()
        data["topic"] = self.topic
        data["atoms"] = []
        data["contexts"] = []

        # Write the atoms
        for aid, atom in self.atoms.items():
            atom_data = dict(
                id=aid, text=atom.get_text(), contexts=list(atom.get_contexts().keys())
            )
            if atom.get_label() is not None:
                atom_data["label"] = atom.get_label()
            data["atoms"].append(atom_data)

        # Write the contexts
        data["contexts"] = [context.to_json() for context in self.contexts.values()]

        # Write to a JSON file (if any)
        if json_file_path:
            with open(json_file_path, "w") as f:
                json.dump(data, f, indent=4)
            f.close()
            print(f"[FactScore] Pipeline instance written to: {json_file_path}")

        return data

    def build(
            self,
            query: str = None,
            response: str = None,
            topic: str = None,
            has_atoms: bool = False,
            has_contexts: bool = False,
            revise_atoms: bool = False,
            use_fast_retriever: bool = True
    ):
        """
        Build the atoms and contexts using the retrieval service.

        Args:
            query: str
                The input user query.
            response: str
                The LLM generated response to the input query.
            topic: str
                The topic of the input query/response.
            has_atoms: bool
                A boolean flag indicating if the atoms have already been created.
            has_contexts: bool
                A boolean flag indicating if the contexts have already been created.
            revise_atoms: bool
                A boolean flag indicating that the atoms need to be decontextualized
                (i.e., pronouns he, she, it, ... replaced by the actual entity)
            use_fast_retriever: bool
                Use a fast multi-threaded context retriever if True, else the
                standard sequential retriever.
        """

        # Initialize the scorer
        if query is not None: 
            self.query = query
        if response is not None:
            self.response = response
        if topic is not None:
            self.topic = topic

        self.revise_atoms = revise_atoms

        # Safety checks
        assert self.atom_extractor is not None, \
            f"The atom extractor must be created."
        assert self.atom_reviser is not None, \
            f"The atom reviser must be created."

        print(f"[FactScore] Building the pipeline ...")
        
        # Build the atoms 
        if has_atoms == False:
            self.atoms = build_atoms(
                response=self.response,
                atom_extractor=self.atom_extractor
            )
            self.revise_atoms = True # revise atoms is newly created
            print(f"[FactScore] Extracted {len(self.atoms)} atoms.")
            for aid in self.atoms.keys():
                print(f"[FactScore] {self.atoms[aid]}")

        # Safety checks
        assert len(self.atoms) > 0, \
            f"The atoms must be initialized before running the pipeline."

        # Revise the atoms
        if self.revise_atoms:
            print(f"[FactScore] Revising the atoms ...")
            assert self.response is not None, f"The atom reviser requires a response."
            atom_ids = [aid for aid in sorted(self.atoms.keys())]
            old_atoms = [self.atoms[aid].get_text() for aid in atom_ids]
            result = asyncio.run(self.atom_reviser.run_batch(old_atoms, self.response))
            for i, aid in enumerate(atom_ids):
                elem = result[i]
                self.atoms[aid].set_text(elem["revised_unit"])
                print(f"[FactScore] {self.atoms[aid]}")

        # Remove duplicated atoms (if any)
        self.atoms = remove_duplicated_atoms(self.atoms)
        print(f"[FactScore] Created {len(self.atoms)} unique atoms.")

        # Build the contexts (per atom)
        if has_contexts == False: # check if contexts already in file
            self.contexts = build_contexts(
                atoms=self.atoms,
                retriever=self.context_retriever,
                use_fast_retriever=use_fast_retriever,
            )
        print(f"[FactScore] Retrieved {len(self.contexts)} contexts.")
        print(f"[FactScore] Pipeline building completed.")

    def _get_label(self, output: ModelOutputThunk) -> str:
        """
        Extract the atom label from the generated text. We expect the label to
        be on the last line of the response, and be one of the following:
            [Supported], [Contradicted], [Unverifiable].
        We only consider [Supported]/S atoms, the others will be [NotSupported]/NS.
        """

        text = str(output)
        generated_answer = text.lower()
        if "true" in generated_answer or "false" in generated_answer:
            if "true" in generated_answer and "false" not in generated_answer:
                is_supported = True
            elif "false" in generated_answer and "true" not in generated_answer:
                is_supported = False
            else:
                is_supported = generated_answer.index("true") > generated_answer.index("false")
        else:
            is_supported = all([keyword not in generated_answer.lower().translate(str.maketrans("", "", string.punctuation)).split() for keyword in ["not", "cannot", "unknown", "information"]])

        label = "S" if is_supported else "NS"
        return label
                
    async def predict_atom_labels(self) -> Tuple[Dict[str, str], Dict[str, str]]:
        """
        For each atom predict its label (S or NS) given the corresponding
        retrieved contexts.
        """

        # Safety checks
        assert len(self.atoms) > 0

        # Utility function to assemble the context of an atom
        def make_knowledge(passages: List[Dict[str, Any]]) -> str:
            knowledge = ""
            for _, psg in enumerate(passages):
                title = psg["title"]
                text = psg["text"]
                snippet = psg.get("snippet", "")
                knowledge += "Title: {}\nSummary: {}\nText: {}\n\n".format(title, snippet, text)

            return knowledge
        
        # Use the LLM to label the atom
        print(f"[FactScore] Labeling atoms with {self.backend.model_id} ...")

        # Label each atom given its retrieved contexts
        atom_ids = []
        atom_labels = []
        atom_outputs = []
        corutines = []
        for aid, atom in self.atoms.items():
            atom_ids.append(aid)
            atom_text = atom.get_text()
            contexts = atom.get_contexts()
            passages = []
            if contexts is not None and len(contexts) > 0:
                for _, c in contexts.items():
                    passages.append(dict(title=c.get_title(), text=c.get_text()))
            
            # Prepare the context
            knowledge_text = make_knowledge(passages)
            print(f"[FactScore] Processing atom: ({aid}) {atom_text}")

            # Prepare the instruction
            if self.topic is not None:
                instruction = INSTRUCTION_FACTSCORE
                user_variables = {
                    "topic_text": self.topic, 
                    "atom_text": atom_text, 
                    "knowledge_text": knowledge_text
                }
            else:
                instruction = INSTRUCTION_FACTSCORE_NOTOPIC
                user_variables = {
                    "atom_text": atom_text, 
                    "knowledge_text": knowledge_text
                }

            # Execute the instruction
            corutine = mfuncs.ainstruct(
                instruction,
                context=SimpleContext(),
                backend=self.backend,
                requirements=[
                    check(
                        "The output must contain the tokens True or False"
                    )
                ],
                user_variables=user_variables,
                strategy=RejectionSamplingStrategy(loop_budget=LOOP_BUDGET),
                return_sampling_results=True
            )
            corutines.append(corutine)
        
        print(f"[FactScore] Awaiting for the async execution ...")
        outputs = await asyncio.gather(*(corutines[i] for i in range(len(corutines))))
        for output in outputs:
            label = self._get_label(output.result)
            atom_labels.append(label)
            atom_outputs.append(str(output))

        # Return the labeled atoms (and also the outputs)
        return dict(zip(atom_ids, atom_labels)), dict(zip(atom_ids, atom_outputs))

    def score(self) -> Dict[str, Any]:
        """
        Compute the factuality score taking into consideration the contexts 
        retrieved for each of the atom in the answer.

        Factuality score = # atoms(true) / # atoms

        Intuitively, a score of 100% means that all atoms in the answer are
        factually correct. If none of them are correct, then the score is 0%. If
        only half of the atoms are correct, then the score is 50%.

        Returns:
            Dict[str, Any]: The results dictionary containing the factuality score i.e., a real value in [0, 1]
        """

        # Run the FactScore pipeline
        num_true_atoms = 0
        num_false_atoms = 0
        num_uniform_atoms = 0
        labels, raw_outputs = asyncio.run(self.predict_atom_labels())
        for _, label in labels.items():
            if self.binary_output:
                if label == "S":
                    num_true_atoms += 1
                else:
                    num_false_atoms += 1
            else:
                if label == "S":
                    num_true_atoms += 1
                elif label == "C":
                    num_false_atoms += 1
                else:
                    num_uniform_atoms += 1
      
        # Precision, R@K and F1@K
        fscore = float(num_true_atoms)/float(len(self.atoms))
        K = int(len(self.atoms) / 2) # K is assumed to be half
        recall_k = min(float(num_true_atoms)/K, 1.0)
        try:
            f1k = 2 * fscore * recall_k / (fscore + recall_k)
        except Exception as _:
            f1k = 0.0

        # Elapsed time
        elapsed_time = time.perf_counter() - self.start_time # elapsed time

        results = {}
        results["factuality_score"] = fscore
        results["recall_k"] = recall_k
        results["f1_k"] = f1k
        results["num_atoms"] = len(self.atoms)
        results["num_contexts"] = len(self.contexts)
        results["num_true_atoms"] = num_true_atoms
        results["num_false_atoms"] = num_false_atoms
        results["num_uniform_atoms"] = num_uniform_atoms
        results["entropy"] = None
        results["avg_entropy"] = None

        print(f"[FactScore] Predictions: {labels}")
        if self.labels_human is not None and self.binary_output is True:
            true_atoms = 0
            false_atoms = 0
            num_true_positive = 0
            num_true_negative = 0
            num_false_positive = 0
            num_false_negative = 0
            for aid, l in self.labels_human.items():
                if l == "S":
                    true_atoms += 1
                    if labels[aid] == "S":
                        num_true_positive += 1
                    else:
                        num_false_negative += 1
                else:
                    false_atoms += 1
                    if labels[aid] == "NS":
                        num_true_negative += 1
                    else:
                        num_false_positive += 1                    
            fscore_gold = true_atoms/len(self.labels_human)
            print(f"[FactScore] Gold labels: {self.labels_human}")
            print(f"[FactScore] Predictions: {labels}")
            print(f"[FactScore] Gold fscore: {fscore_gold} ({true_atoms}/{len(self.labels_human)})")
            results["gold_factuality_score"] = fscore_gold
            results["gold_true_atoms"] = true_atoms
            results["true_positive"] = num_true_positive
            results["true_negative"] = num_true_negative
            results["false_positive"] = num_false_positive
            results["false_negative"] = num_false_negative
        elif self.labels_human is not None and self.binary_output is False:
            true_atoms = 0
            false_atoms = 0
            num_true_positive = 0
            num_true_negative = 0
            num_false_positive = 0
            num_false_negative = 0
            for aid, l in self.labels_human.items(): # true labels are either S or NS
                if l == "S":
                    true_atoms += 1
                    if labels[aid] == "S":
                        num_true_positive += 1
                    else:
                        num_false_negative += 1
                else:
                    false_atoms += 1
                    if labels[aid] in ["C", "U"]:
                        num_true_negative += 1
                    else:
                        num_false_positive += 1     

            fscore_gold = true_atoms/len(self.labels_human)
            print(f"[FactScore] Gold labels: {self.labels_human}")
            print(f"[FactScore] Predictions: {labels}")
            print(f"[FactScore] Gold fscore: {fscore_gold} ({true_atoms}/{len(self.labels_human)})")
            results["gold_factuality_score"] = fscore_gold
            results["gold_true_atoms"] = true_atoms
            results["true_positive"] = num_true_positive
            results["true_negative"] = num_true_negative
            results["false_positive"] = num_false_positive
            results["false_negative"] = num_false_negative

        results["topic"] = self.topic
        results["query"] = self.query
        results["response"] = self.response
        results["elapsed_time"] = elapsed_time
        results["predictions"] = labels
        results["raw_outputs"] = raw_outputs
        print(f"[FactScore] Elapsed time: {elapsed_time:.4f} seconds.")

        return results




"""Core Components for the FactReasoner Library."""

from .base import Atom, Context, Relation
from .atomizer import Atomizer
from .reviser import Reviser
from .retriever import ContextRetriever, Retriever
from .nli import NLIExtractor
from .query_builder import QueryBuilder
from .summarizer import ContextSummarizer
from .utils import (
    build_atoms,
    build_contexts,
    build_relations
)

__all__ = [
    "Atomizer",
    "Reviser",
    "ContextRetriever",
    "Retriever",
    "NLIExtractor",
    "QueryBuilder",
    "ContextSummarizer",
    "Atom",
    "Context",
    "Relation",
    "build_atoms",
    "build_contexts",
    "build_relations",
]


# coding=utf-8
# Copyright 2023-present the International Business Machines.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Main runner script

import os
import json
import argparse
import pandas as pd

from mellea.backends import ModelOption

# Local imports
from fact_reasoner.assessor import FactReasoner
from fact_reasoner.baselines.factscore import FactScore
from fact_reasoner.baselines.factverify import FactVerify
from fact_reasoner.baselines.veriscore import VeriScore
from fact_reasoner.core.atomizer import Atomizer
from fact_reasoner.core.reviser import Reviser
from fact_reasoner.core.retriever import ContextRetriever
from fact_reasoner.core.query_builder import QueryBuilder
from fact_reasoner.core.summarizer import ContextSummarizer
from fact_reasoner.core.nli import NLIExtractor

if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--input_file',
        type=str,
        default=None,
        help="Path to the input dataset (jsonl)."
    )

    parser.add_argument(
        '--output_dir',
        type=str,
        default=None,
        help="Path to the output directory."
    )

    parser.add_argument(
        '--cache_dir',
        type=str,
        default=None,
        help="Path to the cache directory."
    )

    parser.add_argument(
        '--dataset_name',
        type=str,
        default=None,
        help="Name of the dataset."
    )

    parser.add_argument(
        '--service_type',
        type=str,
        default="google",
        help="Service type (wikipedia, chromadb, google)."
    )

    parser.add_argument(
        '--model_id',
        type=str,
        default=None,
        help="Name of the model used by the pipeline."
    )

    parser.add_argument(
        '--pipeline',
        type=str,
        default="factreasoner",
        required=True,
        help="Factuality pipeline (factreasoner, factscore, veriscore, factverify)."
    )

    parser.add_argument(
        '--pipeline_version',
        type=str,
        default="v2",
        help="FactReasoner version: v1, v2 or v3"
    )

    parser.add_argument(
        '--top_k', 
        type=int, 
        default=3, 
        help="Top k results retrieved as contexts per atom."
    )

    parser.add_argument(
        '--use_priors', 
        default=False, 
        action='store_true', 
        help="Use the atom and context priors in the factor definition."
    )

    parser.add_argument(
        '--use_summarizer', 
        default=False, 
        action='store_true', 
        help="Use the ContextSummarizer to summarize contexts (FactReasoner only)."
    )

    parser.add_argument(
        '--use_query_builder', 
        default=False, 
        action='store_true', 
        help="Use the QueryBuilder to generate queries for Google search."
    )

    parser.add_argument(
        '--merlin_path',
        type=str,
        default="/home/radu/git/fm-factual/lib/merlin",
        help="Path to the probabilistic inference engine merlin."
    )

    # Parse the CLI arguments
    args = parser.parse_args()

    # FactReasoner versions:
    if args.pipeline_version == "v1":  # 1 - context-atom relationships only, allow duplicated contexts
        rel_context_context = False
        remove_duplicates = False
        contexts_per_atom_only = True
        option = "1"
    elif args.pipeline_version == "v2":  # 2 - context-atom relationships only, no duplicated contexts
        rel_context_context = False
        remove_duplicates = True
        contexts_per_atom_only = False
        option = "2"
    elif args.pipeline_version == "v3":  # 3 - context-atom and context-context relationships, no duplicated contexts
        rel_context_context = True
        remove_duplicates = True
        contexts_per_atom_only = False
        option = "3"
    else:
        raise ValueError(f"Unknown FactReasoner version: {args.version}")

    # Create the Mellea backend
    from mellea_ibm.rits import RITSBackend, RITS

    # Create a Mellea RITS backend
    if args.model_id == "llama3":
        backend = RITSBackend(
            RITS.LLAMA_3_3_70B_INSTRUCT, 
            model_options={ModelOption.MAX_NEW_TOKENS: 4096}
        )
    elif args.model_id == "granite4":
        backend = RITSBackend(
            RITS.GRANITE_4_H_SMALL, 
            model_options={ModelOption.MAX_NEW_TOKENS: 4096}
        )
    elif args.model_id == "mistral":
        backend = RITSBackend(
            RITS.MISTRAL_LARGE_3_675B_2512, 
            model_options={ModelOption.MAX_NEW_TOKENS: 4096}
        )
    elif args.model_id == "gpt-oss":
        backend = RITSBackend(
            RITS.GPT_OSS_120B,
            model_options={ModelOption.MAX_NEW_TOKENS: 4096}
        )
    else:
        raise ValueError(f"Unknown LLM backend.")

    # from mellea.helpers.fancy_logger import FancyLogger
    # FancyLogger.get_logger().setLevel(FancyLogger.ERROR)

    # Create the atom extractor
    atom_extractor = Atomizer(backend)
    
    # Create the atom reviser
    atom_reviser = Reviser(backend)

    # Create the NLI extractor
    nli_extractor = NLIExtractor(backend)

    # Create the Query Builder
    query_builder = QueryBuilder(backend) if args.use_query_builder else None

    # Create context retriever and summarizer
    context_summarizer = ContextSummarizer(backend)
    context_retriever = ContextRetriever(
        service_type=args.service_type, 
        top_k=args.top_k, 
        cache_dir=args.cache_dir,
        query_builder=query_builder,
        fetch_text=True if args.pipeline != "factverify" else False,
    )

    print(f"Processing input dataset: {args.input_file}")
    filename = args.input_file  # a jsonl file

    # Load the dataset
    with open(filename) as f:
        lines = f.read().splitlines()
    df_inter = pd.DataFrame(lines)
    df_inter.columns = ['json_element']
    df_inter['json_element'].apply(json.loads)
    df = pd.json_normalize(df_inter['json_element'].apply(json.loads))
    dataset = df.to_dict('records')

    print(f"Loading data from: {filename}")
    print(f"Found {len(dataset)} elements")

    # Set the pipeline name
    if args.pipeline in ["factscore", "factverify", "veriscore"]:
        pipeline_name = args.pipeline
    elif args.pipeline == "factreasoner":
        pipeline_name = f"{args.pipeline}-{args.version}"
    else:
        raise ValueError(f"Unknown pipeline: {args.pipeline}. Aborting.")

    # Check if previous results exist. If yes, load them and skip over them
    # when processing the input dataset.
    filename = "eval_{}_{}_{}_{}.jsonl".format(
        pipeline_name,
        args.service_type,
        args.dataset_name,
        args.model_id
    )

    # Prepare the output file
    output_filename = os.path.join(args.output_dir, filename)
    print(f"Reading previous results from: {output_filename}")
    evaluation_data = []
    if os.path.isfile(output_filename):
        with open(output_filename, "r") as f:
            lines = f.readlines()
            for line in lines:
                evaluation_data.append(json.loads(line))

    print(f"Found {len(evaluation_data)} existing evaluations data.")
    print(f"Using factuality pipeline: {pipeline_name}")

    # Loop over the data points in the dataset
    for input_data in dataset:
        # Check if current data has been processed already
        processed = False
        for eval_data in evaluation_data:
            if eval_data["input"] == input_data["input"]:
                processed = True
                break
        if processed:
            prompt = input_data["input"]
            print(f"Input: {prompt} already processed.")
            continue

        # Process the data point with the FactReasoner pipeline
        if args.pipeline == "factreasoner":
            pipeline = FactReasoner(
                atom_extractor=atom_extractor,
                atom_reviser=atom_reviser,
                nli_extractor=nli_extractor,
                context_retriever=context_retriever,
                context_summarizer=context_summarizer,
                merlin_path=args.merlin_path,
                use_priors=args.use_priors
            )
        elif args.pipeline == "factscore":
            pipeline = FactScore(
                backend=backend,
                atom_extractor=atom_extractor,
                atom_reviser=atom_reviser,
                context_retriever=context_retriever,
            )
        elif args.pipeline == "veriscore":
            pipeline = VeriScore(
                backend=backend,
                atom_extractor=atom_extractor,
                atom_reviser=atom_reviser,
                context_retriever=context_retriever,
            )
        elif args.pipeline == "factverify":
            pipeline = FactVerify(
                atom_extractor=atom_extractor,
                atom_reviser=atom_reviser,
                context_retriever=context_retriever,
            )

        # Load the problem instance from a file or dict
        ok = pipeline.from_dict_with_contexts(input_data)
        if not ok:
            continue  # annotations are null (ignore)

        # Build the FactReasoner pipeline
        if args.pipeline == "factreasoner":
            pipeline.build(
                remove_duplicates=remove_duplicates,
                contexts_per_atom_only=contexts_per_atom_only,
                has_atoms=True,
                has_contexts=True,
                revise_atoms=False,
                rel_atom_context=True,
                rel_context_context=rel_context_context,
                summarize_contexts=args.use_summarizer
            )

            results, marginals = pipeline.score()
            results["model_name"] = args.model_id
            evaluation_data.append(results)
            print(f"[FactReasoner] Marginals: {marginals}")
            print(f"[FactReasoner] Results: {results}")
        elif args.pipeline == "factscore":
            pipeline.build(
                has_atoms=True,
                has_contexts=True,
                revise_atoms=False
            )

            # Print the results
            results = pipeline.score()
            results["model_name"] = args.model_id
            evaluation_data.append(results)
            print(f"[FactScore] Results: {results}")
        elif args.pipeline == "veriscore":
            pipeline.build(
                has_atoms=True,
                has_contexts=True,
                revise_atoms=False
            )

            # Print the results
            results = pipeline.score()
            results["model_name"] = args.model_id
            evaluation_data.append(results)
            print(f"[VeriScore] Results: {results}")
        elif args.pipeline == "factverify":
            pipeline.build(
                has_atoms=True,
                has_contexts=True,
                revise_atoms=False
            )

            # Print the results
            results = pipeline.score()
            results["model_name"] = args.model_id
            evaluation_data.append(results)
            print(f"[FactVerify] Results: {results}")

        # Save results to a file
        filename = "eval_{}_{}_{}_{}.jsonl".format(
            pipeline_name,
            args.service_type,
            args.dataset_name,
            args.model_id
        )
        output_filename = os.path.join(args.output_dir, filename)
        print(f"Writing results to: {output_filename}")
        with open(output_filename, "w") as f:
            for res in evaluation_data:
                f.write(f"{json.dumps(res)}\n")

    print("Done.")



# coding=utf-8
# Copyright 2023-present the International Business Machines.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Search API

import os
import json
import sqlite3
import requests
import logging

from ast import literal_eval
from dotenv import load_dotenv
from thefuzz import fuzz

logger = logging.getLogger(__name__)

class SearchAPI():
    def __init__(self, cache_dir: str = None, similarity_threshold: float = 90):
        """ 
        Initialize the SearchAPI with a cache directory and similarity threshold.
        
        Args:
            cache_dir : str
                Path to the SQLite database file for caching search results. If
                it is None, then the retrieved search results are not cached.
            similarity_threshold : float
                Minimum similarity score (0-100) for cached results to be considered relevant.
        """
        if not os.environ.get("_DOTENV_LOADED"):
            load_dotenv(override=True) 
            os.environ["_DOTENV_LOADED"] = "1"

        self.serper_key = os.getenv("SERPER_API_KEY")
        self.url = "https://google.serper.dev/search"
        self.headers = {'X-API-KEY': self.serper_key,
                        'Content-Type': 'application/json'}

        # Set up database if cache_dir is not None
        if cache_dir is not None:
            self.do_caching = True
            self.cache_dir = cache_dir
            os.makedirs(self.cache_dir, exist_ok=True)
            self.database = os.path.join(cache_dir, "my_database.db")
            self.similarity_threshold = similarity_threshold
            self._init_db()
        else:
            self.do_caching = False

    def _init_db(self):
        """
        Initialize the SQLite database with FTS5 for full-text search, using 
        WAL mode for better concurrency.
        """
        assert (self.do_caching is True), f"Caching requires an existing cache dir."

        with sqlite3.connect(self.database) as conn:
            cursor = conn.cursor()
            cursor.execute("PRAGMA journal_mode=WAL;")  # Enable Write-Ahead Logging
            cursor.execute("""
                CREATE VIRTUAL TABLE IF NOT EXISTS search_cache USING fts5(
                    query, response
                )
            """)
            conn.commit()

    def get_snippets(self, claim_lst):
        """
        Retrieve search snippets for a list of claims.
        
        Args:
            claim_lst : list
                A list of claims (strings) for which to retrieve search snippets.
        Returns:
            dict
                A dictionary where keys are claims and values are lists of search results,
                each containing a title, snippet, and link.
        """
        
        text_claim_snippets_dict = {}
        for query in claim_lst:
            search_result = self.get_search_res(query)
            if "statusCode" in search_result:
                logger.error(search_result['message'])
                exit()
            organic_res = search_result.get("organic", [])

            search_res_lst = [{"title": item.get("title", ""),
                               "snippet": item.get("snippet", ""),
                               "link": item.get("link", "")}
                              for item in organic_res]
            text_claim_snippets_dict[query] = search_res_lst
        return text_claim_snippets_dict

    def get_search_res(self, query):
        """
        Retrieve search results for a given query, using cache if available.
        Args:
            query : str
                The search query string.
        Returns:
            dict
                The search results in JSON format, either from cache or from the API.
        """
        
        if self.do_caching:
            cache_key = query.strip()
            cached_response = self._get_from_cache(cache_key)
            if cached_response:
                logger.info(f"CACHE HIT! key={cache_key}, q={cached_response['searchParameters']['q']}")
                return cached_response

        # Make API request
        payload = json.dumps({"q": query})
        response = requests.request("POST", self.url, params={"num": 15}, headers=self.headers, data=payload)
        response_json = literal_eval(response.text)
        if response is not None:
            response.close()
        try:
            if self.do_caching:
                self._save_to_cache(cache_key, response_json)
        except sqlite3.Error as e:
            logger.error(f"Error saving to cache: {e}. Continuing...")
        return response_json

    def _get_from_cache(self, query):
        """
        Retrieve the top 3 most relevant cached search results and use 
        fuzz.token_sort_ratio to pick the best one.
        
        Args:
            query : str
                The search query string to look up in the cache.
        Returns:
            dict or None
                The best matching cached search result as a dictionary, or None if no match is found.
        """

        # Escape double quotes for FTS5
        query = query.replace('"', '""')

        with sqlite3.connect(self.database) as conn:
            cursor = conn.cursor()
            cursor.execute("""
                SELECT query, response FROM search_cache
                WHERE query MATCH ?
                LIMIT 3
            """, (f'"{query}"',))
            rows = cursor.fetchall()

        if not rows:
            return None

        # Compute similarity scores using fuzz.token_sort_ratio
        best_match = max(rows, key=lambda row: fuzz.token_sort_ratio(query, row[0]))
        best_score = fuzz.token_sort_ratio(query, best_match[0])

        return json.loads(best_match[1]) if best_score > self.similarity_threshold else None

    def _save_to_cache(self, query, response_json):
        """
        Save a search result to the SQLite FTS5 cache with transactions.
        
        Args:
            query : str
                The search query string to cache.
            response_json : dict
                The search result in JSON format to cache.
        Returns:
            None
        Raises:
            sqlite3.Error: If there is an error during the database operation.
        """
        
        # Do not cache empty search results
        organic_res = response_json.get("organic", [])
        if len(organic_res) == 0:
            return
        
        conn = sqlite3.connect(self.database)
        try:
            cursor = conn.cursor()
            cursor.execute("BEGIN TRANSACTION;")  # Start transaction
            cursor.execute("REPLACE INTO search_cache (query, response) VALUES (?, ?)",
                           (query, json.dumps(response_json)))
            conn.commit()  # Commit transaction
        except sqlite3.Error as e:
            conn.rollback()  # Rollback on error
            logger.error(f"Database error: {e}")
        finally:
            conn.close()


if __name__ == '__main__':
    
    cache_dir = None # "my_database.db"

    text = "Neil B. Todd was an American geneticist"
    # text = "Lanny Flaherty is an American."
    # text = "\"Benefits of displaying menu prices without tax included\" OR \"Advantages of separate tax display in restaurant menus\""
    # text = text.replace('"', '')

    # Initialize search api
    web_search = SearchAPI(cache_dir=cache_dir)

    claim_lst = [text]
    claim_snippets = web_search.get_snippets(claim_lst)
    print(claim_snippets)
    print("Done.")



# coding=utf-8
# Copyright 2023-present the International Business Machines.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# NLI extractor using LLMs.

import asyncio
import math
import mellea.stdlib.functional as mfuncs

from typing import Any, Dict, List

from mellea.backends import Backend
from mellea.stdlib.context import SimpleContext
from mellea.core import ModelOutputThunk
from mellea.stdlib.requirements import check, simple_validate
from mellea.stdlib.sampling import RejectionSamplingStrategy
from mellea.core import FancyLogger

# Local imports
from fact_reasoner.utils import extract_last_square_brackets

INSTRUCTION_NLI = """

Instructions:
You are provided with a PREMISE and a HYPOTHESIS. 
Your task is to evaluate the relationship between the PREMISE and the HYPOTHESIS, following the steps outlined below:

1. Evaluate Relationship:
- If the PREMISE strongly implies or directly supports the HYPOTHESIS, explain the supporting evidence.
- If the PREMISE contradicts the HYPOTHESIS, identify and explain the conflicting evidence.
- If the PREMISE is insufficient to confirm or deny the HYPOTHESIS, explain why the evidence is inconclusive.
2. Provide the reasoning behind your evaluation of the relationship between PREMISE and HYPOTHESIS, justifying each decision.
3. Final Answer: Based on your reasoning, the HYPOTHESIS and the PREMISE, determine your final answer. \
Your final answer must be one of the following: entailment, contradiction or neutral, wrapped in square brackets:
- [entailment] if the PREMISE strongly implies, directly supports or entails the HYPOTHESIS.
- [contradiction] if the PREMISE contradicts the HYPOTHESIS.
- [neutral] if the PREMISE and the HYPOTHESIS neither entail nor contradict each other.

Use the following examples to better understand your task.

Example 1:
PREMISE: Robert Haldane Smith, Baron Smith of Kelvin, KT, CH, FRSGS is a British businessman and former Governor of the British Broadcasting Corporation. Smith was knighted in 1999, appointed to the House of Lords as an independent crossbench peer in 2008, and appointed Knight of the Thistle in the 2014 New Year Honours.
HYPOTHESIS: Robert Smith holds the title of Baron Smith of Kelvin.
1. Evaluate Relationship:
The PREMISE states that Robert Haldane Smith, Baron Smith of Kelvin, KT, CH, FRSGS is a British businessman and former Governor of the British Broadcasting Corporation. It also mentions that Smith was appointed to the House of Lords as an independent crossbench peer in 2008. This information directly supports the HYPOTHESIS that Robert Smith holds the title of Baron Smith of Kelvin.
2: Reasoning:
The PREMISE explicitly mentions that Robert Smith is Baron Smith of Kelvin, which directly supports the HYPOTHESIS. The additional information about his knighthood, appointment to the House of Lords, and other titles further confirms his status as a peer, but it is not necessary to support the specific HYPOTHESIS about him holding the title of Baron Smith of Kelvin.
3. Final Answer: 
[entailment]

Example 2:
PREMISE: In 2022, Passover begins in Israel at sunset on Friday, 15 April, and ends at sunset on Friday, 22 April 2022.
HYPOTHESIS: Passover in 2022 begins at sundown on March 27.
1. Evaluate Relationship:
The PREMISE states that Passover in 2022 begins at sunset on Friday, 15 April, and ends at sunset on Friday, 22 April 2022. The HYPOTHESIS claims that Passover in 2022 begins at sundown on March 27. 
Upon analyzing the information, I found that the dates mentioned in the PREMISE and the HYPOTHESIS do not match. Since the dates provided in the PREMISE and the HYPOTHESIS are different, the HYPOTHESIS is contradicted by the PREMISE.
2. Reasoning:
The PREMISE provides specific information about the start date of Passover in 2022, which is April 15. The HYPOTHESIS, on the other hand, claims a different start date, March 27. This discrepancy indicates that the PREMISE and the HYPOTHESIS cannot both be true.
3. Final Answer:
[contradiction]

Example 3:
PREMISE: Little India in the East Village: Two restaurants ablaze with tiny colored lights stand at the top of a steep staircase.
HYPOTHESIS: The village had colorful decorations on every street corner.
1. Evaluate Relationship:
The PREMISE describes a specific scene in Little India in the East Village, where two restaurants are decorated with tiny colored lights at the top of a steep staircase. The HYPOTHESIS makes a broader claim that the village had colorful decorations on every street corner.
The PREMISE provides evidence of colorful decorations in one specific location, but it does not provide information about the decorations on every street corner in the village. The PREMISE is insufficient to confirm or deny the HYPOTHESIS, as it only describes a small part of the village.
2. Reasoning:
The PREMISE and HYPOTHESIS are related in that they both mention colorful decorations, but the scope of the HYPOTHESIS is much broader than the PREMISE. The PREMISE only provides a glimpse into one specific location, whereas the HYPOTHESIS makes a general claim about the entire village. Without more information, it is impossible to determine whether the village had colorful decorations on every street corner.
3. Final Answer:
[neutral]

Your task:
PREMISE: {{premise_text}}
HYPOTHESIS: {{hypothesis_text}}
"""

class NLIExtractor:
    """
    Predict the NLI relationship between a premise and a hypothesis, optionally
    given a context (or response). The considered relationships are: entailment,
    contradiction and neutrality. We use few-shot prompting for LLMs.

    v1 - original
    v2 - more recent (with reasoning)
    v3 - only for Google search results
    """
    
    def __init__(
            self,
            backend: Backend,
    ):
        """
        Initialize the NLIExtractor.

        Args:
            backend: Backend
                The Mellea backend to use for LLM interaction.
        """

        # Safety checks        
        if backend is None:
            raise ValueError("Mellea backend is None. Please provide a valid Mellea backend.")

        self.method = "logprobs"
        self.backend = backend

        # Print info
        print(f"[NLI] Using Mellea backend: {self.backend.model_id}")

        # Disable Mellea logging
        FancyLogger.get_logger().setLevel(FancyLogger.ERROR)


    def _get_probability(self, output: ModelOutputThunk) -> float:
        """
        Compute the average log probability of the generated tokens.

        Args:
            output: ModelOutputThunk
                The model raw output (via Mellea).

        Returns:
            float: The average log probability of the generated tokens.
        """

        assert output._meta["oai_chat_response"]["logprobs"] is not None
        logprobs = output._meta["oai_chat_response"]["logprobs"]["content"][:-1] # last token is EOS

        # Go backwards and collect the logprobs of the tokens between ']' and ']'
        avg_logprob = 0
        count = 0
        for item in reversed(logprobs):
            if item['token'] == '[':
                break
            elif item['token'] == ']':
                continue
            else:
                avg_logprob += item['logprob']
                count += 1

        # Compute the probability
        avg_logprob = avg_logprob / count if count > 0 else math.inf
        return math.exp(avg_logprob) if not math.isinf(avg_logprob) else 0.0 

    def _get_label(self, output: ModelOutputThunk) -> str:
        """
        Extract the NLI label from the model output.

        Args:
            output: ModelOutputThunk
                The model raw output (via Mellea)

        Returns:
            str: The string representing the NLI label (entailment, contradiction, neutral).        
        """

        return extract_last_square_brackets(str(output))

    def run(self, premise: str, hypothesis: str) -> Dict[str, Any]:
        """
        Extract the NLI relationship between premise and hypothesis. The 
        following relationships are allowed: entailment, contradiction, neutral.
        
        Args:
            premise: str
                The premise text (e.g., context).
            hypothesis: str
                The hypothesis text (e.g., atom).

        Returns:
            Dict[str, Any]: A dictionary containing the relationship and its probability.
        """
        
        # Perform the instruction with validation
        output = mfuncs.instruct(
            INSTRUCTION_NLI,
            context=SimpleContext(),
            backend=self.backend,
            requirements=[
                check(
                    "The output must be a wrapped in square brackets",
                    validation_fn=simple_validate(
                        lambda s: extract_last_square_brackets(s) != ''
                    ),
                )
            ],
            user_variables={"premise_text": premise, "hypothesis_text": hypothesis},
            strategy=RejectionSamplingStrategy(loop_budget=3),
            return_sampling_results=True,
            model_options=dict(logprobs=True),
        )

        if output.success:
            label = self._get_label(output.result)
            probability = self._get_probability(output.result)
            if label not in ["entailment", "contradiction", "neutral"]:
                label = "neutral"

            return dict(
                label=label, 
                probability=probability
            )
        else:
            return dict(
                label="neutral", 
                probability=1.0
            )

    async def run_batch(self, premises: List[str], hypotheses: List[str]) -> List[Dict[str, Any]]:
        """
        Extract the NLI relationships between premises and hypotheses. The 
        following relationships are allowed: entailment, contradiction, neutral.
        
        Args:
            premises: List[str]
                The list of premise texts (e.g., context).
            hypotheses: List[str]
                The list of hypothesis texts (e.g., atom).

        Returns:
            List[Dict[str, Any]]: A list of dictionaries containing the 
            relationships and their probabilities.
        """

        corutines = []
        for premise, hypothesis in zip(premises, hypotheses):
            corutine = mfuncs.ainstruct(
                INSTRUCTION_NLI,
                context=SimpleContext(),
                backend=self.backend,
                requirements=[
                    check(
                        "The output must be a wrapped in square brackets",
                        validation_fn=simple_validate(
                            lambda s: extract_last_square_brackets(s) != ''
                        ),
                    )
                ],
                user_variables={"premise_text": premise, "hypothesis_text": hypothesis},
                strategy=RejectionSamplingStrategy(loop_budget=3),
                return_sampling_results=True,
                model_options=dict(logprobs=True),
            )
            corutines.append(corutine)

        results = []
        print(f"[NLI] Awaiting for async execution ...")
        outputs = await asyncio.gather(*(corutines[i] for i in range(len(corutines))))
        for output in outputs:
            if output.success:
                label = self._get_label(output.result)
                probability = self._get_probability(output.result)
                results.append(dict(
                    label=label,
                    probability=probability
                ))
            else:
                results.append(dict(
                    label="neutral",
                    probability=1.0
                ))

        return results




# coding=utf-8
# Copyright 2023-present the International Business Machines.g
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Optional, Union

# Defaut prior probabilities for atoms and contexts
PRIOR_PROB_ATOM = 0.5
PRIOR_PROB_CONTEXT = 0.9

class Atom:
    """
    Represents an atomic unit of the model's response.
    """

    def __init__(
            self,
            id: str,
            text: str,
            label: str = None
    ):
        """
        Atom constructor.
        Args:
            id: str
                Unique ID of the atom e.g., `a1`.
            text: int
                The text associated with the atom.
            label: str
                The gold label associated with the atom (S or NS).
        """

        self.id = id
        self.text = text
        self.original = text  # keeps around the original atom
        self.label = label
        self.contexts = {}
        self.search_results = []
        self.probability = PRIOR_PROB_ATOM  # prior probability of the atom being true

    def __str__(self) -> str:
        return f"Atom {self.id}: {self.text}"

    def get_text(self):
        return self.text
    
    def get_summary(self):
        return self.text

    def set_text(self, new_text: str):
        self.text = new_text

    def get_original(self):
        return self.original

    def set_original(self, new_original: str):
        self.original = new_original

    def get_label(self):
        return self.label

    def add_context(
            self,
            context
    ):
        """
        Add a context relevat to the atom.

        Args:
            context: Context
                The context relevat to the atom.
        """
        self.contexts[context.id] = context

    def add_contexts(
            self,
            contexts
    ):
        """
        Add a list of contexts relevant to the atom.
        Args:
            context: list
                The contexts relevant to the atom.
        """
        for context in contexts:
            self.contexts[context.id] = context

    def get_contexts(self):
        """
        Return the contexts relevant to the atom.
        """
        return self.contexts


class Context:
    """
    Represents a context retrieved from an external source of knowledge.
    """

    def __init__(
            self,
            id: str,
            atom: Optional[Atom],
            text: str = "",
            synthetic_summary: Optional[str] = None,
            title: str = "",
            link: str = "",
            snippet: str = ""
    ):
        """
        Context constructor.
        Args:
            id: str
                Unique ID for the context e.g., `c1_1`.
            atom: Atom
                The reference atom (from the answer)
            text: str
                The text of the context (one or more paragraphs)
            sumary: str
                The summary of the context (one or more paragraphs)
            title: str
                The title of the context (e.g., title of the wikipedia page)
            link: str
                The link to a web page if the context is a search results. It
                is assumed to be empty if the context is a retrieved passage.
            snippet: str
                The snippet associated with a search result
        """

        self.id = id
        self.atom = atom
        self.text = text
        self.synthetic_summary = synthetic_summary
        self.title = title
        self.link = link
        self.snippet = snippet
        self.probability = PRIOR_PROB_CONTEXT  # prior probability of the context being true

    def __str__(self) -> str:
        return f"Context {self.id} [{self.title}]: {self.text}"

    def get_id(self):
        return self.id

    def get_summary(self):
        return self.synthetic_summary if self.synthetic_summary is not None else ""

    def get_text(self):
        if self.snippet != "" and self.text != "":
            return "Snippet/Summary of Text:\n\n" + self.snippet + "\n\n" + "Text:\n\n" + self.text 
        elif self.snippet == "" and self.text != "":
            return self.text
        elif self.snippet != "" and self.text == "":
            return self.snippet
        else:
            return ""
       
    def get_title(self):
        return self.title

    def get_link(self):
        return self.link

    def get_snippet(self):
        return self.snippet

    def set_atom(self, atom):
        self.atom = atom

    def set_link(self, link: str):
        self.link = link

    def set_snippet(self, snippet: str):
        self.snippet = snippet

    def set_synthetic_summary(self, synthetic_summary: str):
        self.synthetic_summary = synthetic_summary

    def get_probability(self):
        return self.probability

    def set_probability(self, probability):
        self.probability = probability

    def to_json(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "title": self.title,
            "text": self.text,
            "link": self.link,
            "snippet": self.snippet,
            "synthetic_summary": self.get_summary(),
            "probability": self.probability,
        }     


class Relation:
    """
    Represents the NLI relationship between a source text and a target text.
    """

    def __init__(
            self,
            source: Union[Atom, Context],
            target: Union[Atom, Context],
            type: str,
            probability: float,
            link: str
    ):
        """
        Relation constructor.
        Args:
            source: [Atom|Context]
                The source atom or context.
            target: [Atom|Context]
                The target atom or context.
            type: str
                The relation type: ["entailment", "contradiction", "equivalence"].
                Note that `entailment` is not symmetric while `contradiction` and
                `equivalence` are symmetric relations.
            probability: float
                The probability value associated with the NLI relation.
            link: str
                The link type: [context_atom, context_context, atom_atom]
        
        Comment: `entailment` is not symmetric, while `contradiction`, `neutral`
        and `equivalence are symmetric. Namely, if A contradicts B then B
        contradicts A (same with neutral, equivalence). However, if A entails B
        then B doesn't necessarily entails A.
        """

        if "entailment" in type.lower():
            type = "entailment"
        elif "contradiction" in type.lower():
            type = "contradiction"
        elif "equivalence" in type.lower():
            type = "equivalence"
        elif "neutral" in type.lower():
            type = "neutral"
        else:
            type = "neutral"  # default to neutral if the type is not recognized
            probability = 1.0  # set probability to 1 for neutral relations
            
        assert (link in ["context_atom", "context_context", "atom_atom"]), \
            f"Unknown link type: {link}"

        self.source = source
        self.target = target
        self.type = type
        self.probability = probability
        self.link = link

    def __str__(self) -> str:
        return f"[{self.source.id} -> {self.target.id}] : {self.type} : {self.probability}"

    def get_type(self) -> str:
        return self.type

    def get_probability(self) -> float:
        return self.probability

